import inspect
import warnings
from typing import Any, Dict, List, Mapping, Optional

from fastapi import Query
from monty.json import MSONable
from pydantic import BaseModel
from pydantic.fields import ModelField

from maggma.api.util import STORE_PARAMS, dynamic_import
from maggma.core import Store


class QueryOperator(MSONable):
    """
    Base Query Operator class for defining powerfull query language
    in the Materials API
    """

    def query(self) -> STORE_PARAMS:
        """
        The query function that does the work for this query operator
        """
        raise NotImplementedError("Query operators must implement query")

    def meta(self, store: Store, query: Dict) -> Dict:
        """
        Returns meta data to return with the Response

        Args:
            store: the Maggma Store that the resource uses
            query: the query being executed in this API call
        """
        return {}

    def post_process(self, doc: Dict) -> Dict:
        """
        An optional post-processing function for the data
        """
        return doc


class PaginationQuery(QueryOperator):
    """Query opertators to provides Pagination in the Materials API"""

    def __init__(
        self, default_skip: int = 0, default_limit: int = 10, max_limit: int = 100
    ):
        """
        Args:
            default_skip: the default number of documents to skip
            default_limit: the default number of documents to return
            max_limit: max number of documents to return
        """
        self.default_skip = default_skip
        self.default_limit = default_limit
        self.max_limit = max_limit

        def query(
            skip: int = Query(
                default_skip, description="Number of entries to skip in the search"
            ),
            limit: int = Query(
                default_limit,
                description="Max number of entries to return in a single query."
                f" Limited to {max_limit}",
            ),
        ) -> STORE_PARAMS:
            """
            Pagination parameters for the API Endpoint
            """
            if limit > max_limit:
                raise Exception(
                    "Requested more data per query than allowed by this endpoint."
                    f"The max limit is {max_limit} entries"
                )
            return {"skip": skip, "limit": limit}

        self.query = query  # type: ignore

    def meta(self, store: Store, query: Dict) -> Dict:
        """
        Metadata for the pagination params
        """
        return {"max_limit": self.max_limit}


class SparseFieldsQuery(QueryOperator):
    def __init__(self, model: BaseModel, default_fields: Optional[List[str]] = None):
        """
        Args:
            model: PyDantic Model that represents the underlying data source
            default_fields: default fields to return in the API response if no fields are explicitly requested
        """

        self.model = model

        model_fields = list(self.model.__fields__.keys())
        self.default_fields = (
            model_fields if default_fields is None else list(default_fields)
        )

        assert set(self.default_fields).issubset(
            model_fields
        ), "default projection contains some fields that are not in the model fields"

        default_fields_string = ",".join(self.default_fields)  # type: ignore

        def query(
            fields: str = Query(
                default_fields_string,
                description=f"Fields to project from {model.__name__} "  # type: ignore
                f"as a list of comma seperated strings",
            ),
            all_fields: bool = Query(False, description="Include all fields."),
        ) -> STORE_PARAMS:
            """
            Projection parameters for the API Endpoint
            """

            projection_fields: List[str] = [s.strip() for s in fields.split(",")]

            # need to strip to avoid input such as "name, weight" to be parsed into ["name", " weight"]
            # we need ["name", "weight"]
            if all_fields:
                projection_fields = model_fields
            return {"properties": projection_fields}

        self.query = query  # type: ignore

    """
    Factory function to generate a dependency for sparse field sets in FastAPI
    """

    def meta(self, store: Store, query: Dict) -> Dict:
        """
        Returns metadata for the Sparse field set
        """
        return {"default_fields": self.default_fields}

    def as_dict(self) -> Dict:
        """
        Special as_dict implemented to convert pydantic models into strings
        """

        d = super().as_dict()  # Ensures sub-classes serialize correctly
        d["model"] = f"{self.model.__module__}.{self.model.__name__}"  # type: ignore
        return d

    @classmethod
    def from_dict(cls, d):
        """
        Special from_dict to autoload the pydantic model from the location string
        """
        model = d.get("model")
        if isinstance(model, str):
            module_path = ".".join(model.split(".")[:-1])
            class_name = model.split(".")[-1]
            model = dynamic_import(module_path, class_name)

        assert issubclass(
            model, BaseModel
        ), "The resource model has to be a PyDantic Model"
        d["model"] = model

        cls(**d)


class DefaultDynamicQuery(QueryOperator):
    def __init__(
        self,
        model: BaseModel,
        additional_signature_fields: Mapping[str, List] = None,
    ):
        """
        This function will take query, parse it, and output the mongo criteria.

        About the query:
        The format of the input query will come from two sources: the data model (which we will sometimes refer to as
        default values, and the additional paremeters that users can choose to pass in)
        additional_signature_field must be int the shape of NAME_OF_THE_QUERY -> [DEFAULT_VALUE, QUERY]
        where QUERY is a FastAPI params.Query object

        About how does this script parse the query:
        it will first generate default queries from input data model
        Then it will merge with the optional additional_signature_fields

        About mongo criteria output:
        current implementation only supports 6 operations, namely equal, not equal, less than, greater than, in,
        and not in. Therefore, the criteria output will be also limited to those operations.


        Args:
            model: PyDantic Model to base the query language on
            additional_signature_fields: mapping of NAME_OF_THE_FIELD -> [DEFAULT_VALUE, QUERY]
        """
        self.model = model
        default_mapping = {
            "eq": "$eq",
            "not_eq": "$ne",
            "lt": "$lt",
            "gt": "$gt",
            "in": "$in",
            "not_in": "$nin",
        }
        mapping: dict = default_mapping

        self.additional_signature_fields: Mapping[str, List] = (
            dict()
            if additional_signature_fields is None
            else additional_signature_fields
        )

        # construct fields
        # find all fields in data_object
        all_fields: Dict[str, ModelField] = model.__fields__

        # turn fields into operators, also do type checking
        params = self.fields_to_operator(all_fields)

        # combine with additional_fields
        # user's input always have higher priority than the the default data model's
        params.update(self.additional_signature_fields)

        def query(**kwargs) -> STORE_PARAMS:
            crit = dict()
            for k, v in kwargs.items():
                if v is not None:
                    if "_" not in k:
                        k = k + "_eq"
                    name, operator = k.split("_", 1)
                    try:
                        crit[name] = {mapping[operator]: v}
                    except KeyError:
                        raise KeyError(
                            f"Cannot find key {k} in current query to database mapping"
                        )

            return {"criteria": crit}

        # building the signatures for FastAPI Swagger UI
        signatures: List[Any] = []
        signatures.extend(
            inspect.Parameter(
                param,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=query[1],
                annotation=query[0],
            )
            for param, query in params.items()
        )
        setattr(query, "__signature__", inspect.Signature(signatures))

        self.query = query  # type: ignore

    def fields_to_operator(self, all_fields: Dict[str, ModelField]) -> Dict[str, list]:
        """
        Getting a list of tuple of fields

        turn them into a mapping of

        FIELD_[operator] -> query


        Args:
            all_fields: dictionary of str -> ModelField

        Returns:
            a dictionary of FIELD_[operator] -> query
        """
        params = dict()
        for name, model_field in all_fields.items():
            if model_field.type_ in [str, int, float]:
                t: Any = model_field.type_
                name = model_field.name
                params[f"{name}"] = [
                    t,
                    Query(
                        default=model_field.default,
                        description=f"Querying if {model_field.name} is equal to another",
                    ),
                ]  # supporting both with and with out explicit _eq
                params[f"{name}_eq"] = [
                    t,
                    Query(
                        default=model_field.default,
                        description=f"Querying if {model_field.name} is equal to another",
                    ),
                ]
                params[f"{name}_not_eq"] = [
                    t,
                    Query(
                        default=model_field.default,
                        description=f"Querying if {model_field.name} is not equal to another",
                    ),
                ]
                params[f"{name}_in"] = [
                    List[t],
                    Query(
                        default=model_field.default,
                        description=f"Querying if item is in {model_field.name}",
                    ),
                ]
                params[f"{name}_not_in"] = [
                    List[t],
                    Query(
                        default=model_field.default,
                        description=f"Querying if item is not in {model_field.name} ",
                    ),
                ]

                if model_field.type_ == int or model_field == float:
                    params[f"{name}_lt"] = [
                        model_field.type_,
                        Query(
                            model_field.default,
                            description=f"Querying if {model_field.name} is less than to another",
                        ),
                    ]
                    params[f"{name}_gt"] = [
                        model_field.type_,
                        Query(
                            model_field.default,
                            description=f"Querying if {model_field.name} is greater than to another",
                        ),
                    ]
            else:
                warnings.warn(
                    f"Field name {name} with {model_field.type_} not implemented"
                )
        return params



from random import randint
from urllib.parse import urlencode

import pytest
from fastapi import FastAPI
from pydantic import BaseModel, Field
from requests import Response
from starlette.testclient import TestClient

from maggma.api.resource import Resource
from maggma.stores import MemoryStore


class Owner(BaseModel):
    name: str = Field(..., title="Owner's name")
    age: int = Field(None, title="Owne'r Age")
    weight: float = Field(None, title="Owner's weight")


@pytest.fixture(scope="session")
def owners():
    return (
        [
            Owner(
                name=f"Person{i}", age=randint(10, 100), weight=float(randint(100, 100))
            )
            for i in list(range(10)[1:])
        ]  # there are 8 owners here
        + [Owner(name="PersonAge9", age=9, weight=float(randint(155, 195)))]
        + [Owner(name="PersonWeight150", age=randint(10, 15), weight=float(150))]
        + [Owner(name="PersonAge20Weight200", age=20, weight=float(200))]
    )


@pytest.fixture
def owner_store(owners):
    store = MemoryStore("owners", key="name")
    store.connect()
    owners = [d.dict() for d in owners]
    store.update(owners)
    return store


def test_init_resource(owner_store):
    num_default_api_routes = 2
    endpoint = Resource(owner_store, Owner)
    assert len(endpoint.router.routes) == num_default_api_routes
    assert endpoint.router.routes[0]

    endpoint = Resource(owner_store, "tests.api.test_resource.Owner")
    assert len(endpoint.router.routes) == num_default_api_routes
    assert endpoint.router.routes[0]


def test_resource_msonable(owner_store):
    endpoint = Resource(owner_store, Owner)
    endpoint_dict = endpoint.as_dict()

    for k in ["@class", "@module", "store", "model"]:
        assert k in endpoint_dict

    assert isinstance(endpoint_dict["model"], str)
    assert endpoint_dict["model"] == "tests.api.test_resource.Owner"


def test_resource_get_by_key(owner_store):
    endpoint = Resource(owner_store, Owner)
    app = FastAPI()
    app.include_router(endpoint.router)

    client = TestClient(app)

    assert client.get("/").status_code == 200

    assert client.get("/Person1/").status_code == 200
    assert client.get("/Person1/").json()["data"][0]["name"] == "Person1"


def dynamic_model_search_response_helper(
    client, payload, base: str = "/?", debug=False
) -> Response:
    """
    Helper function for dynamicModelSearch function, to reduce redundant code
    Args:
        base: base of the query, default to /query?
        client: TestClient generated from FastAPI
        payload: query in dictionary format
        debug: True = print out the url, false don't print anything

    Returns:
        request.Response object that contains the response of the correspoding payload
    """
    url = base + urlencode(payload)
    if debug:
        print(url)
    res = client.get(url)
    return res


def get_fields(res: Response):
    json = res.json()
    return json.get("data", []), json.get("meta", {}), json.get("error", [])


def test_resource_dynamic_model_search_eq_noteq(owner_store):
    endpoint = Resource(owner_store, Owner)
    app = FastAPI()
    app.include_router(endpoint.router)

    client = TestClient(app)
    payload = {"name_eq": "PersonAge9", "all_fields": True}
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?"
    )
    data, meta, err = get_fields(res)

    # String eq
    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["name"] == "PersonAge9"

    # int Eq
    assert data[0]["age"] == 9

    # float Eq
    payload = {"weight_eq": 150.0, "all_fields": True}
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?"
    )
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["weight"] == 150.0

    # Str Not eq
    payload = {"name_not_eq": "PersonAge9", "all_fields": True}
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?"
    )
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 10

    # int Not Eq
    payload = {"age_not_eq": 9, "all_fields": True}
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?"
    )
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 10

    # float Not Eq
    payload = {"weight_not_eq": 150.0, "all_fields": True}
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?"
    )
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 10


def test_resource_dynamic_model_search_lt_gt(owner_store):
    endpoint = Resource(owner_store, Owner)
    app = FastAPI()
    app.include_router(endpoint.router)

    client = TestClient(app)
    payload = {
        "age_lt": 101  # see if there are 10 entries, all age should be less than 100
    }

    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?"
    )
    data, meta, err = get_fields(res)

    assert res.status_code == 200
    assert len(data) == 10

    payload = {"age_gt": 0}  # there should be 10 entries, all age are greater than 10
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?"
    )
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 10


def test_resource_dynamic_model_search_in_notin(owner_store):
    endpoint = Resource(owner_store, Owner)
    app = FastAPI()
    app.include_router(endpoint.router)

    client = TestClient(app)

    # In
    url = "/?age_in=9&age_in=101&all_fields=true&limit=10"
    res = client.get(url)
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 1

    # Not In
    url = "/?limit=10&all_fields=true&age_not_in=9&age_not_in=10"
    res = client.get(url)
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 10


def test_resource_no_explicit_eq(owner_store):
    endpoint = Resource(owner_store, Owner)
    app = FastAPI()
    app.include_router(endpoint.router)

    client = TestClient(app)

    payload = {"name": "PersonAge9", "all_fields": True}
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?", debug=True
    )
    data, meta, err = get_fields(res)

    # String eq with no explicit eq
    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["name"] == "PersonAge9"


def test_resource_compound(owner_store):
    endpoint = Resource(owner_store, Owner)
    app = FastAPI()
    app.include_router(endpoint.router)

    client = TestClient(app)

    payload = {
        "name": "PersonAge20Weight200",
        "all_fields": True,
        "weight": 200,
        "age": 20,
    }
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?", debug=True
    )
    data, meta, err = get_fields(res)

    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["name"] == "PersonAge20Weight200"

    payload = {
        "name": "PersonAge20Weight200",
        "all_fields": False,
        "fields": "name, age",
        "weight": 200,
        "age": 20,
    }
    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/?", debug=True
    )
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["name"] == "PersonAge20Weight200"
    assert data[0]["age"] == 20
    assert "weight" not in data[0]



from inspect import isclass
from typing import Dict

import uvicorn
from fastapi import FastAPI
from monty.json import MSONable

from maggma.api.resource import Resource
from maggma.api.util import dynamic_import


class APIManager(MSONable):
    def __init__(self, endpoints: Dict[str, Resource]):
        self.endpoints = endpoints

    @property
    def app(self):
        """
        App server for the cluster manager
        """
        app = FastAPI()
        if len(self) == 0:
            raise RuntimeError("ERROR: There are no endpoints provided")

        for prefix, endpoint in self.endpoints.items():
            app.include_router(endpoint.router, prefix=f"/{prefix}")
        return app

    def run(self, ip: str = "127.0.0.1", port: int = 8000, log_level: str = "info"):
        """
        Runs the Cluster Manager locally

        Args:
            ip: Local IP to listen on
            port: Local port to listen on
            log_level: Logging level for the webserver

        Returns:
            None
        """
        uvicorn.run(self.app, host=ip, port=port, log_level=log_level, reload=False)

    def load(self, endpoint, prefix: str = "/"):
        """
        loads an endpoint dynamically. The endpoint can be either a path to an EndpointCluster instance,
        or a EndpointCluster instance Args: endpoint:

        Returns:
            None

        Raises:
            ValueError -- if the endpoint is not a path to an EndpointCluster or it is not an EndpointCluster
        """
        if isinstance(endpoint, str):
            module_path = ".".join(endpoint.split(".")[:-1])
            class_name = endpoint.split(".")[-1]
            new_endpoint = dynamic_import(module_path, class_name)
            self.__setitem__(prefix, new_endpoint)

        elif isclass(endpoint) and issubclass(endpoint, Resource):
            self.__setitem__(prefix, endpoint)
        else:
            raise ValueError(
                "endpont has to be a EndpointCluster instance or a path to EndpointCluster instance"
            )

    def __setitem__(self, key, item):
        self.endpoints[key] = item

    def __getitem__(self, key):
        return self.endpoints[key]

    def __len__(self):
        return len(self.endpoints)

    def keys(self):
        return self.endpoints.keys()

    def __contains__(self, item):
        return item in self.endpoints

    def sort(self):
        """
        Return :
            A sorted list of endpoints sorted in lexicographical order.

            This will gaurentee that paths with shorter path are before longer paths

        """
        return sorted(list(self.endpoints.keys()))

    def pprint(self):
        """
        pretty print an list of endpoints in the following order:
        1. By lexicographical
        2. By inheritance
        Example: if we have ["/temp", "/main","/main/temp"], this function would return the list in the order of
        ["/main","/main/temp","/temp"]


        """
        for line in self.sort():
            print(line)



from enum import Enum
from random import choice, randint
from urllib.parse import urlencode

import pytest
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel, Field
from requests import Response
from starlette.testclient import TestClient

from maggma.api.APIManager import APIManager
from maggma.api.resource import Resource
from maggma.stores import MemoryStore


class PetType(str, Enum):
    cat = "cat"
    dog = "dog"


class Owner(BaseModel):
    name: str = Field(..., title="Owner's name")
    age: int = Field(..., title="Owne'r Age")
    weight: int = Field(..., title="Owner's weight")


class Pet(BaseModel):
    name: str = Field(..., title="Pet's Name")
    pet_type: PetType = Field(..., title="Pet Type")
    owner_name: str = Field(..., title="Owner's name")


@pytest.fixture(scope="session")
def owners():
    return [
        Owner(name=f"Person{i}", age=randint(10, 100), weight=randint(100, 200))
        for i in list(range(10)[1:])
    ]


@pytest.fixture(scope="session")
def pets(owners):
    return [
        Pet(
            name=f"Pet{i}",
            pet_type=choice(list(PetType)),
            owner_name=choice(owners).name,
        )
        for i in list(range(40))[1:]
    ]


@pytest.fixture
def owner_store(owners):
    store = MemoryStore("owners", key="name")
    store.connect()
    owners = [jsonable_encoder(d) for d in owners]
    store.update(owners)
    return store


@pytest.fixture
def pet_store(pets):
    store = MemoryStore("pets", key="name")
    store.connect()
    pets = [jsonable_encoder(d) for d in pets]
    store.update(pets)
    return store


def test_cluster_dict_behavior(owner_store, pet_store):
    owner_endpoint = Resource(owner_store, Owner)
    pet_endpoint = Resource(pet_store, Pet)

    manager = APIManager({"owners": owner_endpoint})

    assert manager["owners"] == owner_endpoint  # test __getitem__
    assert len(manager) == 1  # test len
    assert "owners" in manager.keys()  # test keys
    assert "owners" in manager  # test __contains__
    assert "pets" not in manager.keys()  # test __keys__

    manager["pets"] = pet_endpoint  # test __setitem__
    assert len(manager) == 2
    assert manager["pets"] == pet_endpoint


def dynamic_model_search_response_helper(
    client, payload, base: str = "/search?"
) -> Response:
    """
    Helper function for dynamicModelSearch function, to reduce redundant code
    Args:
        base: base of the query, default to /query?
        client: TestClient generated from FastAPI
        payload: query in dictionary format

    Returns:
        request.Response object that contains the response of the correspoding payload
    """
    url = base + urlencode(payload)
    print(url)
    res = client.get(url)
    return res


def get_fields(res: Response):
    json = res.json()
    return json.get("data", []), json.get("meta", {}), json.get("error", [])


def test_cluster_run(owner_store, pet_store):
    owner_endpoint = Resource(owner_store, Owner)
    pet_endpoint = Resource(pet_store, Pet)

    manager = APIManager({"owners": owner_endpoint, "pets": pet_endpoint})

    client = TestClient(manager.app)

    assert client.get("/").status_code == 404
    payload = {"name_eq": "Person1", "limit": 10, "all_fields": True}

    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/owners/?"
    )
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["name"] == "Person1"

    payload = {"name_eq": "Pet1", "limit": 10, "all_fields": True}

    res = dynamic_model_search_response_helper(
        client=client, payload=payload, base="/pets/?"
    )
    data, meta, err = get_fields(res)
    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["name"] == "Pet1"


def test_cluster_pprint(owner_store, pet_store):
    endpoint_main = Resource(owner_store, Owner, description="main")
    endpoint_main_temp = Resource(pet_store, Pet, description="main_temp")
    endpoint_temp = Resource(owner_store, Owner, description="temp")

    manager = APIManager(
        {
            "/temp": endpoint_temp,
            "/main": endpoint_main,
            "/main/temp": endpoint_main_temp,
        }
    )

    res = manager.sort()

    assert res == ["/main", "/main/temp", "/temp"]



from typing import Any, Dict, List, Optional, Union

from fastapi import APIRouter, Depends, FastAPI, HTTPException, Path
from monty.json import MSONable
from pydantic import BaseModel

from maggma.api.models import Meta, Response
from maggma.api.query_operator import (
    DefaultDynamicQuery,
    PaginationQuery,
    QueryOperator,
    SparseFieldsQuery,
)
from maggma.api.util import (
    STORE_PARAMS,
    attach_signature,
    dynamic_import,
    merge_queries,
)
from maggma.core import Store


class Resource(MSONable):
    """
    Implements a REST Compatible Resource as a URL endpoint
    This class provides a number of convenience features
    including full pagination, field projection, and the
    MAPI query lanaugage
    """

    def __init__(
        self,
        store: Store,
        model: Union[BaseModel, str],
        tags: Optional[List[str]] = None,
        query_operators: Optional[List[QueryOperator]] = None,
        description: str = None,
    ):
        """
        Args:
            store: The Maggma Store to get data from
            model: the pydantic model to apply to the documents from the Store
                This can be a string with a full python path to a model or
                an actuall pydantic Model if this is being instantied in python
                code. Serializing this via Monty will autoconvert the pydantic model
                into a python path string
            tags: list of tags for the Endpoint
            query_operators: operators for the query language
            description: an explanation of wht does this resource do
        """
        self.store = store
        self.tags = tags or []
        self.description = description
        self.model: BaseModel = BaseModel()
        if isinstance(model, str):
            module_path = ".".join(model.split(".")[:-1])
            class_name = model.split(".")[-1]
            self.model = dynamic_import(module_path, class_name)
        else:
            self.model = model

        self.query_operators = (
            query_operators
            if query_operators is not None
            else [
                PaginationQuery(),
                SparseFieldsQuery(self.model, default_fields=[self.store.key]),
                DefaultDynamicQuery(self.model),
            ]
        )

        self.response_model = Response[self.model]  # type: ignore
        self.router = APIRouter()
        self.prepare_endpoint()

    def prepare_endpoint(self):
        """
        Internal method to prepare the endpoint by setting up default handlers
        for routes
        """
        self.build_get_by_key()
        self.set_dynamic_model_search()

    def build_get_by_key(self):
        key_name = self.store.key
        model_name = self.model.__name__

        async def get_by_key(
            key: str = Path(
                ..., alias=key_name, title=f"The {key_name} of the {model_name} to get"
            ),
            fields: STORE_PARAMS = Depends(
                SparseFieldsQuery(self.model, [self.store.key]).query
            ),
        ):
            f"""
            Get's a document by the primary key in the store

            Args:
                {key_name}: the id of a single {model_name}

            Returns:
                a single {model_name} document
            """
            self.store.connect()

            item = self.store.query_one(
                criteria={self.store.key: key}, properties=fields["properties"]
            )

            if item is None:
                raise HTTPException(
                    status_code=404,
                    detail=f"Item with {self.store.key} = {key} not found",
                )

            response = {"data": [item]}  # , "meta": Meta()}
            return response

        self.router.get(
            f"/{{{key_name}}}/",
            response_description=f"Get an {model_name} by {key_name}",
            response_model=self.response_model,
            response_model_exclude_unset=True,
            tags=self.tags,
        )(get_by_key)

    def set_dynamic_model_search(self):

        model_name = self.model.__name__

        async def search(**queries: STORE_PARAMS):
            self.store.connect()

            query: Dict[Any, Any] = merge_queries(list(queries.values()))

            count_query = query["criteria"]
            count = self.store.count(count_query)
            data = list(self.store.query(**query))
            meta = Meta(total=count)
            response = {"data": data, "meta": meta.dict()}
            return response

        attach_signature(
            search,
            annotations={
                f"dep{i}": STORE_PARAMS for i, _ in enumerate(self.query_operators)
            },
            defaults={
                f"dep{i}": Depends(dep.query)
                for i, dep in enumerate(self.query_operators)
            },
        )

        self.router.get(
            "/",
            tags=self.tags,
            summary=f"Get {model_name} documents",
            response_model=self.response_model,
            response_description=f"Search for a {model_name}",
            response_model_exclude_unset=True,
        )(search)

    def run(self):  # pragma: no cover
        """
        Runs the Endpoint cluster locally
        This is intended for testing not production
        """
        import uvicorn

        app = FastAPI()
        app.include_router(self.router, prefix="")
        uvicorn.run(app)

    def as_dict(self) -> Dict:
        """
        Special as_dict implemented to convert pydantic models into strings
        """

        d = super().as_dict()  # Ensures sub-classes serialize correctly
        d["model"] = f"{self.model.__module__}.{self.model.__name__}"  # type: ignore
        return d
